import numpy as np
import scipy.linalg as la

check_finite = True

def cho_inv(A):
    L,low = la.cho_factor(A, check_finite=check_finite)
    B = np.eye(L.shape[0])
    B = la.cho_solve((L,low),B,overwrite_b=True,check_finite=check_finite)
    return B

def cho_solve(A,B):
    L,low = la.cho_factor(A, check_finite=check_finite)
    C = la.cho_solve((L,low),B,overwrite_b=False, check_finite=check_finite)
    return C

class EKF:
    """
    class for extended kalman filter providing (hopefully) user-friendly API
    """
    def __init__(self, nstate: int, nmeas: int, *, 
                    dtype: np.dtype = np.float64, delta: float = 1e-3):
        """Initialise EKF for a given dimension of system.
        """
        self._nstate  = nstate
        self._nmeas   = nmeas
        self._dtype   = dtype
        self._delta   = delta
        self._init_ss = False
        self._static_jacobian = None
        self._jacobian_func   = None
        self.x_k = np.zeros([self._nstate],self._dtype)
        self.P_k = np.zeros([self._nstate,self._nstate],self._dtype)
        self.K_k = np.zeros([self._nstate,self._nmeas],self._dtype)
        self.H_k = np.zeros([self._nmeas,self._nstate],self._dtype)

    def set_h_model(self,h_model,*args):
        """Set non-linear function h(x).
        h(x) must be a function of one vector argument, with length=nstate, and it must
        return a vector output, with length=nmeas.
        """
        self._h_model = h_model
        self._h_args  = args
        assert self._h_model(np.zeros(self._nstate,dtype=self._dtype),
                                *self._h_args).shape == (self._nmeas,)

    def update(self,e_k,use_static_jacobian=False):
        # perform upate
        self.P_k = self.SigV + self.Amat @ self.P_k @ self.Amat.T
        if use_static_jacobian:
            if self._static_jacobian is None:
                self._set_static_jacobian()
            self.H_k = self._static_jacobian
        else:
            self.H_k = self.linearise_h(self.x_k)
        self.K_k = np.linalg.solve(self.H_k @ self.P_k @ self.H_k.T + self.SigW,
                            self.H_k @ self.P_k.T).T
        self.x_k = self.Amat @ self.x_k + self.K_k @ e_k
        self.P_k = self.P_k - self.P_k @ self.H_k.T @ self.K_k.T    

    def update_ss(self,e_k):
        # perform steady state upate
        if not self._init_ss:
            self.init_steady_state()
        self.x_k = self.Amat @ self.x_k + self.K_ss @ e_k

    def init_steady_state(self,epochs=50):
        if self._static_jacobian is None:
            self._set_static_jacobian()
        self.H_ss = self._static_jacobian
        self.P_ss,KT_ss = self.DARE(self.Amat.T,self.H_k.T,self.SigV,self.SigW,epochs=epochs)
        self.K_ss = KT_ss.T
        self._init_ss = True
    
    def DARE(self,A,B,Q,R,epochs=50):
        alpha   = np.copy(A)
        beta    = np.copy(B@np.linalg.solve(R,B.T))
        gamma   = np.copy(Q)
        Id = np.eye(Q.shape[0])
        print("%d/%d Complete" % (0, epochs))
        old = alpha
        for it in range(epochs):
            common  = Id + beta@gamma
            gamma   = gamma + alpha.T@gamma@np.linalg.solve(common,alpha)
            beta    = beta + alpha@np.linalg.solve(common,beta)@alpha.T
            alpha   = alpha@np.linalg.solve(common,alpha)
            print("%d/%d Complete" % (it+1, epochs))
            diff = np.sum(np.abs(alpha.flatten()-old.flatten()))
            print("difference: %f" % diff)
            old     = alpha.copy()
        return [gamma,np.linalg.solve(R+B.T@gamma@B,B.T)@gamma@A]

    def _set_static_jacobian(self,x0=None):
        if x0 is None:
            x0 = 0.0*self.x_k
        self._static_jacobian = self.linearise_h(x0)

    def linearise_h(self,r):
        dr = np.zeros_like(r)
        dr[0] = self._delta
        y_0 = self._h_model(r,*self._h_args)
        y_len = y_0.shape[0]
        jacobian = np.zeros([y_len,dr.shape[0]],dtype=self._dtype)
        for j in range(dr.shape[0]):
            y_tmp = self._h_model(r+np.roll(dr,j),*self._h_args)
            jacobian[:,j] = (y_tmp-y_0) / self._delta
        return jacobian

    def set_matrices(self,Amat=None,SigV=None,SigW=None,SigW_inv=None):
        if Amat is not None:
            self.Amat = Amat.copy()
        if SigV is not None:
            self.SigV = SigV.copy()
        if SigW is not None:
            self.SigW = SigW.copy()
        if SigW_inv is not None:
            self.SigW_inv = SigW_inv.copy()

    def set_mats_from_scalars(self,*,alpha=0.995,sigsqx=1e-9,sigsqw=1e-1):
        Amat =  alpha * np.eye(self._nstate,dtype=self._dtype)
        SigX = sigsqx * np.eye(self._nstate,dtype=self._dtype)
        SigV = SigX - Amat @ SigX @ Amat.T
        SigW = sigsqw * np.eye(self._nmeas,dtype=self._dtype)
        SigW_inv = (1/sigsqw) * np.eye(self._nmeas,dtype=self._dtype)
        self.set_matrices(Amat=Amat,SigV=SigV,SigW=SigW,SigW_inv=SigW_inv)

    def reset(self):
        self.P_k = self.SigV.copy()
        self.H_k *= 0.0
        self.K_k *= 0.0
        self.x_k *= 0.0
        self._static_jacobian = None
    
    def set_jacobian_func(self,func):
        self._jacobian_func = func


class IteratedEKF(EKF):
    def update(self,meas_k):
        # perform upate
        self.P_k = self.SigV + self.Amat @ self.P_k @ self.Amat.T
        x_i = self.Amat @ self.x_k

        for it in range(10):
            #print(x_i)
            if self._jacobian_func is not None:
                H_i = self._jacobian_func(x_i)
            else:
                H_i = self.linearise_h(x_i)
            K_i = cho_solve(H_i @ self.P_k @ H_i.T + self.SigW,
                            H_i @ self.P_k.T).T
            x_i = self.x_k + K_i @ (meas_k-self._h_model(x_i)-H_i@(self.x_k-x_i))
        
        self.x_k = x_i 
        self.P_k = (np.eye(x_i.shape[0])-K_i @ H_i) @ self.P_k


try:
    import cupy as cp
except ImportError as ie:
    Warning("No cupy available, using basic numpy")
    import numpy as cp

import numpy as np

import scipy.optimize as opt
from scipy.linalg import block_diag

class FastAndFurious:
    """Fast and Furious is a class for computing the phase of a wavefront
    from an image using sequential phase diversity. It is based on the
    algorithm described in Korkiakoski et al, Appl. Opt. 53, 4565-4579 (2014).

    Parameters
    ----------
    pup_small : ndarray
        The pupil function.
    im_width : int
        The width of the image.
    fft_width : int
        The width of the FFT.
    offset : ndarray (pup_width x pup_width)
        The offset of the wavefront phase (in radians) required to (e.g.) centre
        the focal plane image over 2x2 pixels (rather than centred over 1x1).
    epsilon : float/ndarray (fft_width x fft_width)
        The regularisation parameter used in the focal plane of the FF algorithm. 
        Can be a scalar or an array in the fft domain. E.g., this parameter
        should be set to a large number if the corresponding pixel is noisy/unused.
        See example for details.

    Attributes
    ----------
    pup_width : int
        The width of the pupil plane arrays.
    im_width : int
        The width of the focal plane arrays.
    fft_width : int 
        The width of the FFT support.
    
    Methods
    -------
    get_phase(p_i)
        Compute the phase of the wavefront from an image.
    set_diversity_phase(dphi)
        Set the diversity phase.
    """
    def __init__(self, pup: np.ndarray, im_width: int, fft_width: int, 
                offset: np.ndarray, epsilon: float = 1e-5):
        """ Initialise an instance of FastAndFurious.
        """
        self.im_width   = im_width           # width of image
        self.pup_width  = pup.shape[0] # width of the pupil
        self.fft_width  = fft_width          # width of fft

        # Set up the (hidden) parameters. These are all either scalars or cupy
        # arrays.
        # Pupil in FFT dimensions:
        self._pup = cp.pad(cp.array(pup),(self.fft_width-pup.shape[0])//2) 
        # Offset for (e.g.) centering the image on 2x2 pixels:
        self._offset = self._pup * cp.exp(1j*cp.pad(cp.array(offset),(self.fft_width-self.pup_width)//2))
        # Recursive parameters:
        self._v_d = None
        self._y_d = None
        self._p_de = None
        # Reset flag:
        self._reset = True
        # Focal plane complex amplitude (real valued -- assuming symmetry in pupil):
        self._a = (cp.fft.fftshift(cp.fft.fft2(self._pup*self._offset))).real
        self._a /= self._pup.sum()
        # Focal plane intensity:
        self._a2 = self._a**2
        # Regularisation parameter:
        self._epsilon = cp.array(epsilon)
        # Constants:
        self._ydenom = 2*self._a**2+self._epsilon
        self._S = 1 # TODO: estimate this on the fly      

    def reset(self):
        """ Reset the FF algorithm/recursive parameters
        """
        self._v_d   = None
        self._y_d   = None
        self._p_de  = None
        self._reset = True

    def retrieve_phase(self,p_i):
        """ Compute the phase of the wavefront from an image.

        Note that the phase in all variables within FF is expected to be in
        radians, so typically one will need to scale returned phase by 
        lambda/(2*pi) after calling this function.

        Parameters
        ----------
        p_i : ndarray (im_width x im_width)
            The image used for wavefront retrieval.

        Returns
        -------
        ndarray (pup_width x pup_width)
            The estimated phase of the wavefront.
        """

        # rescale the image and pad to the fft size:
        p_i = cp.pad(cp.array(p_i),(self.fft_width-self.im_width)//2)
        p_i *= self._a2.sum() / p_i.sum()
        p_i += (1-p_i.max()/self._a2.max())*self._a2

        # odd/even decomposition:
        p_io = 0.5*(p_i-p_i[::-1,::-1]) # odd component of image
        p_ie = 0.5*(p_i+p_i[::-1,::-1]) # even component of image
        
        # compute imaginary part of FFT(phase):
        y_i = self._a*p_io/self._ydenom
        
        # compute sign of real part of FFT(phase):
        if self._reset:
            v_sign = cp.sign(self._a.copy())
            self._reset = False
        else:
            v_sign = cp.sign((self._p_de-p_ie-(self._v_d**2+self._y_d**2+2*y_i*self._y_d))/(2*self._v_d))
        
        # compute abs(real part of FFT(phase)):
        v_i_abs = cp.sqrt(cp.abs(p_ie-(self._S*self._a2+y_i**2)))

        # combine to get real part of FFT(phase):
        v_i = v_i_abs * v_sign

        # inverse transform to get phase + negative because I might have a sign 
        # error somewhere else?:
        phi = - cp.fft.ifft2(cp.fft.fftshift(v_i+1j*y_i),norm="forward").real
        
        # save even part of image for next iteration:
        self._p_de = p_ie.copy()

        # return phase estimate on pupil plane support:
        phi =  phi[self.fft_width//2-self.pup_width//2:self.fft_width//2+self.pup_width//2,
                   self.fft_width//2-self.pup_width//2:self.fft_width//2+self.pup_width//2]
        # return numpy array regardless of if using cupy or not:
        if hasattr(phi,"get"):
            return phi.get()
        else:
            return phi

    def set_diversity_phase(self,dphi):
        """ Set the diversity phase. This should be the difference between the 
        most recent phase and the phase of the wavefront at the previous image.
        I.e., the difference between the two DM shapes, assuming the turbulence
        change is negligble.

        Note that the phase in all variables within FF is expected to be in
        radians, so typically one will need to scale the phase by 2*pi/lambda 
        before passing to this function.

        Parameters
        ----------
        dphi : ndarray (pup_width x pup_width)
            The diversity phase. 
        """
        dphi_big = cp.pad(cp.array(dphi),(self.fft_width-self.pup_width)//2)
        dphi_big_o = (dphi_big-dphi_big[::-1,::-1])*0.5
        dphi_big_e = (dphi_big+dphi_big[::-1,::-1])*0.5
        self._y_d = cp.fft.fftshift(cp.fft.ifft2(dphi_big_o)).imag
        self._v_d = cp.fft.fftshift(cp.fft.ifft2(dphi_big_e)).real

class MHEStatic:
    """Moving Horizon Estimator with a static phase assumption (i.e., A=eye)
    """

    def __init__(self, nstate: int, nmeas: int, nbuffer: int, noise_cov: np.ndarray,
            state_cov: np.ndarray, h_eval: callable, h_jac: callable = None, 
            h_hess: callable = None, opt_method: str = 'BFGS',
            cost_scaling: float = 1.0, callback: callable = None):
        """Initialise the estimator
        """
        self._nstate = nstate               # len(x)
        self._nmeas = nmeas                 # len(y)
        self._nbuffer = nbuffer             # mhe horizon length
        
        self._h_eval = h_eval               # h(x)
        self._h_jac = h_jac                 # dh(x)/dx
        self._h_hess = h_hess               # d^2h(x)/dx^2
        self._opt_method = opt_method       # optimisation method

        self._gamma = np.linalg.inv(state_cov)
        self._noise_cov_inv = np.linalg.inv(noise_cov)

        self._cost_scaling = cost_scaling
        self._callback = callback

    def cost(self, x, x_dm, yd):
        """Evaluate MHE cost function for state, x, measurements, yd, and
        dm sequence, x_dm.
        """
        cost  = (x).T @ self._gamma @ (x)
        h  = np.r_[[self._h_eval(x+(x_dm)[self._nstate*i:self._nstate*(i+1)]) 
                            for i in range(self._nbuffer)]]
        cost += np.sum([hi.T @ self._noise_cov_inv @ hi for hi in h],axis=0)
        cost += - 2 * np.sum([hi.T @ self._noise_cov_inv @ ydi 
                            for hi,ydi in zip(h,yd)],axis=0)
        return cost*self._cost_scaling


    def jac(self, x, x_dm, yd):
        """
        x should be (N,NSTATE)
        """
        h     = [self._h_eval(x+(x_dm)[self._nstate*i:self._nstate*(i+1)]) 
                            for i in range(self._nbuffer)]
        dhdx  = [self._h_jac(x+(x_dm)[self._nstate*i:self._nstate*(i+1)]) 
                            for i in range(self._nbuffer)]
        djdx  = 2*(self._gamma @ x)
        djdx += 2*np.sum([dhidx.T @ self._noise_cov_inv @ hi 
                            for dhidx,hi in zip(dhdx,h)],axis=0)
        djdx -= 2*np.sum([dhidx.T @ self._noise_cov_inv @ ydi 
                            for dhidx,ydi in zip(dhdx,yd)],axis=0)
        return djdx*self._cost_scaling

    def hess(self, x, x_dm, yd):
        """
        x should be (N,NSTATE)
        """
        h       = [self._h_eval(x+(x_dm)[self._nstate*i:self._nstate*(i+1)]) 
                            for i in range(self._nbuffer)]
        dhdx    = [self._h_jac(x+(x_dm)[self._nstate*i:self._nstate*(i+1)]) 
                            for i in range(self._nbuffer)]
        d2hdx2  = [self._h_hess(x+(x_dm)[self._nstate*i:self._nstate*(i+1)]) 
                            for i in range(self._nbuffer)]
        d2jdx2  = 2*(self._gamma)
        d2jdx2 += 2*np.sum([d2hidx2.T @ self._noise_cov_inv @ hi 
                            for d2hidx2,hi in zip(d2hdx2,h)],axis=0)
        d2jdx2 += 2*np.sum([dhidx.T @ self._noise_cov_inv @ dhidx 
                            for dhidx in dhdx],axis=0)
        d2jdx2 -= 2*np.sum([d2hidx2.T @ self._noise_cov_inv @ ydi 
                            for d2hidx2,ydi in zip(d2hdx2,yd)],axis=0)
        return d2jdx2*self._cost_scaling
    
    def get_estimate(self, x0, x_dm, yd):
        """Get the current estimate of the state based on recent priors.

        Args:
            x0 (np.ndarray): Initial state estimate (best guess)
            x_dm (np.ndarray): dm sequence (N,NSTATE)
            yd (np.ndarray): Measurement sequence (N,NMEAS)

        Returns:

        """

        cost_fun = lambda x : self.cost(x, x_dm, yd)
        jac_fun  = None
        hess_fun = None        
        if self._h_jac is not None:
            jac_fun = lambda x: self.jac(x, x_dm, yd)
            if self._h_hess is not None:
                hess_fun = lambda x: self.hess(x, x_dm, yd)
        xopt = opt.minimize(cost_fun, x0, jac=jac_fun, hess=hess_fun,
                             method=self._opt_method, callback=self._callback)
        self._xopt = xopt # save most recent output for debugging
        return xopt["x"]

class MHE:
    """Moving Horizon Estimator
    """

    def __init__(self, nstate: int, nmeas: int, nbuffer: int, noise_cov: np.ndarray,
            state_cov: np.ndarray, state_matrix: np.ndarray, h_eval: callable, 
            h_jac: callable = None, h_hess: callable = None, opt_method: str = 'BFGS',
            cost_scaling: float = 1.0, callback: callable = None, opt_options: dict = None):
        """Initialise the estimator
        """
        self._nstate = nstate               # len(x)
        self._nmeas = nmeas                 # len(y)
        self._nbuffer = nbuffer             # mhe horizon length

        self._h_eval = h_eval               # h(x)
        self._h_jac = h_jac                 # dh(x)/dx
        self._h_hess = h_hess               # d^2h(x)/dx^2
        self._opt_method = opt_method       # optimisation method
        self._opt_options = opt_options     # optimisation options
        process_cov = state_cov - state_matrix @ state_cov @ state_matrix.T

        state_cov_inv   = np.linalg.inv(state_cov)
        process_cov_inv = np.linalg.inv(process_cov)
        
        gamma = np.zeros([nstate*nbuffer,nstate*nbuffer])
        gamma[:nstate,:nstate]   = state_cov_inv - process_cov_inv
        gamma[-nstate:,-nstate:] = -state_matrix.T @ process_cov_inv @ state_matrix
        for mi in range(nbuffer):
            for ni in range(nbuffer):
                if mi==ni:
                    tmp = state_matrix.T @ process_cov_inv @ state_matrix + process_cov_inv
                    gamma[mi*nstate:(mi+1)*nstate,ni*nstate:(ni+1)*nstate] += tmp
                elif mi==(ni+1):
                    tmp = -process_cov_inv @ state_matrix
                    gamma[mi*nstate:(mi+1)*nstate,ni*nstate:(ni+1)*nstate] = tmp
                elif mi==(ni-1):
                    tmp = -state_matrix.T @ process_cov_inv
                    gamma[mi*nstate:(mi+1)*nstate,ni*nstate:(ni+1)*nstate] = tmp

        self._gamma = gamma
        self._noise_cov_inv = np.linalg.inv(noise_cov)

        self._cost_scaling = cost_scaling
        self._callback = callback

    def cost(self, x, x_dm, yd):
        """Evaluate MHE cost function for state, x, measurements, yd, and
        dm sequence, x_dm.
        """
        cost  = (x).T @ self._gamma @ (x)
        h  = np.r_[[self._h_eval((x+x_dm)[self._nstate*i:self._nstate*(i+1)]) 
                            for i in range(self._nbuffer)]]
        cost += np.sum([hi.T @ self._noise_cov_inv @ hi for hi in h],axis=0)
        cost += - 2 * np.sum([hi.T @ self._noise_cov_inv @ ydi 
                            for hi,ydi in zip(h,yd)],axis=0)
        return cost*self._cost_scaling


    def jac(self, x, x_dm, yd):
        """
        x should be (N,NSTATE)
        """
        h     = [self._h_eval((x+x_dm)[self._nstate*i:self._nstate*(i+1)]) 
                            for i in range(self._nbuffer)]
        dhdx  = [self._h_jac((x+x_dm)[self._nstate*i:self._nstate*(i+1)]) 
                            for i in range(self._nbuffer)]
        djdx  = 2*(self._gamma @ x)
        djdx += 2*np.concatenate([dhidx.T @ self._noise_cov_inv @ hi 
                            for dhidx,hi in zip(dhdx,h)],axis=0)
        djdx -= 2*np.concatenate([dhidx.T @ self._noise_cov_inv @ ydi 
                            for dhidx,ydi in zip(dhdx,yd)],axis=0)
        return djdx*self._cost_scaling

    def hess(self, x, x_dm, yd):
        """
        x should be (N,NSTATE)
        """
        h       = [self._h_eval((x+x_dm)[self._nstate*i:self._nstate*(i+1)]) 
                            for i in range(self._nbuffer)]
        dhdx    = [self._h_jac((x+x_dm)[self._nstate*i:self._nstate*(i+1)]) 
                            for i in range(self._nbuffer)]
        d2hdx2  = [self._h_hess((x+x_dm)[self._nstate*i:self._nstate*(i+1)]) 
                            for i in range(self._nbuffer)]
        d2jdx2  = 2*(self._gamma)
        d2jdx2 += 2*block_diag(*[d2hidx2.T @ self._noise_cov_inv @ hi 
                            for d2hidx2,hi in zip(d2hdx2,h)])
        d2jdx2 += 2*block_diag(*[dhidx.T @ self._noise_cov_inv @ dhidx 
                            for dhidx in dhdx])
        d2jdx2 -= 2*block_diag(*[d2hidx2.T @ self._noise_cov_inv @ ydi 
                            for d2hidx2,ydi in zip(d2hdx2,yd)])
        return d2jdx2*self._cost_scaling
    
    def get_estimate(self, x0, x_dm, yd):
        """Get the current estimate of the state based on recent priors.

        Args:
            x0 (np.ndarray): Initial state estimate (best guess)
            x_dm (np.ndarray): dm sequence (N,NSTATE)
            yd (np.ndarray): Measurement sequence (N,NMEAS)

        Returns:
            np.ndarray: Current state estimate
        """

        cost_fun = lambda x : self.cost(x, x_dm, yd)
        jac_fun  = None
        hess_fun = None        
        if self._h_jac is not None:
            jac_fun = lambda x: self.jac(x, x_dm, yd)
            if self._h_hess is not None:
                hess_fun = lambda x: self.hess(x, x_dm, yd)
        xopt = opt.minimize(cost_fun, x0, jac=jac_fun, hess=hess_fun,
                             method=self._opt_method, callback=self._callback,
                             options=self._opt_options)
        self._xopt = xopt # save most recent output for debugging
        return xopt["x"][-self._nstate:]

class GerchbergSaxton:
    """Gerchberg-Saxton algorithm for estimating the phase of a wavefront given
    the pupil mask and focal-plane image.

    Args:
        pup : (np.ndarray)
            Pupil mask, will be padded to be fft_width x fft_width.
        wavelength : (float)
            Wavelength of light, in whatever units you like. Retrieved phase will
            be in the same units.
        fft_width : (int)
            Width of the FFT support in pixels.
        im_width : (int)
            Width of the image in pixels.
    
    Methods:
        compute_phase : Compute the phase of the wavefront.

    WARNING: Incomplete implementation.
    Requires:
     - binning considerations.
    """
    def __init__(self,pup,wavelength,fft_width,im_width,offset=None):
        self._im_width = im_width
        self._fft_width = fft_width
        self._pup_width = pup.shape[0]
        pup_big = cp.pad(cp.array(pup),(self._fft_width-pup.shape[0])//2)
        self._pup_shft = cp.fft.fftshift(pup_big)
        self._wavelength = wavelength
        self._scf = (cp.abs(cp.fft.fft2(self._pup_shft))**2).sum()
        self._offset_shft = cp.fft.fftshift(cp.pad(cp.array(offset),(self._fft_width-offset.shape[0])//2))
        #self.half_pix_phase = cp.mgrid[:self.fft_width,:self.fft_width].sum(axis=0)*(2*cp.pi/self.fft_width/2)
        #self.half_pix_phase = self.half_pix_phase[self.pup==1]
        self.invalid_pixels = cp.fft.fftshift(
                cp.pad(cp.ones([self._im_width,self._im_width]),
                (self._fft_width-self._im_width)//2) == 0)

    def _rebin(self, a, newshape ):
        """
        Rebin an array to a new shape.
        
        Args:
            a (np.ndarray): Array to be rebinned
            newshape (tuple): New shape of the array

        Returns:
            np.ndarray: Rebinned array
        """
        slices = [ slice(None,None, old/new) for old,new in zip(a.shape,newshape) ]
        return a[slices]

    def _gs_run(self,phasepup_shft,amp_shft,iterations=20,hio_param=0.3,invalid_pixels=None):
        """Run the GS algorithm on the input values.

        Args:
            phasepup_shft : (cp.ndarray)
                Initial guess of phase at pupil after fft-shifting (fft_width,fft_width).
            amp_shft : (cp.ndarray)
                Focal-plane amplitude after fft-shifting (fft_width,fft_width).
            iterations : (int)
                Number of iterations to run.
            hio_param : (float)
                Parameter for HIO step.
            invalid_pixels : (cp.ndarray)
                Invalid pixels (fft_width,fft_width). If not None, will be used to
                mask out invalid pixels in focal-plane.

        Returns:
            phasepup_shft : (cp.ndarray)
                Phase estimate from GS iterations (im_width,im_width)
        """
        
        amppup_shft = self._pup_shft.copy()
        for it in range(iterations):
            cplxpup_shft = ((1.+hio_param)*self._pup_shft-hio_param*amppup_shft)*cp.exp(1j*phasepup_shft)    
            cplxim_shft = cp.fft.fft2(cplxpup_shft)
            
            phaseim_shft = cp.angle(cplxim_shft)
            ampim_shft = cp.abs(cplxim_shft)
            
            cplxim_shft2 = ((1.+hio_param)*amp_shft-hio_param*ampim_shft)*cp.exp(1j*phaseim_shft)
            if invalid_pixels is not None:
                cplxim_shft2[invalid_pixels] = cplxim_shft[invalid_pixels].copy() 

            cplxpup_shft2 = cp.fft.ifft2(cplxim_shft2)
            phasepup_shft = cp.angle(cplxpup_shft2)
            amppup_shft = cp.abs(cplxpup_shft2)
            
        return phasepup_shft

    def compute_phase(self,im,iters=100,init=None,hio_param=0.3,discard_invalid=True):
        """Compute and return the phase estimated by the Gerchberg-Saxton algorithm,
        given an image.

        Args:
            im (np.ndarray): Image to perform GS on.
            iters (int): Number of iterations of GS to perform.
            init (np/cp.ndarray): Initial guess for the phase.
            hio_param (float): Parameter for the HIO step.
            discard_invalid (bool): If True, discard invalid pixels in WFS image.

        Returns:
            np.ndarray: Phase estimate.
        """
        if init is np.ndarray:
            init = cp.array(init)
        im = cp.array(im)*(self._scf/im.sum())
        im = cp.pad(im,(self._fft_width-self._im_width)//2)
        im = im**0.5
        im_shft = cp.fft.fftshift(im)
        
        if init is None:
            init = cp.random.randn(*self._pup_shft.shape)
        else:
            if cp.all(cp.r_[init.shape]==self._fft_width):
                pass
            else:
                init = cp.pad(init,(self._fft_width-init.shape[0])//2)
        if discard_invalid:
            invalid_pixels = self.invalid_pixels
        else:
            invalid_pixels = None
        out_phi_shft = self._gs_run(init,im_shft,iterations=iters,hio_param=hio_param,
                                    invalid_pixels=invalid_pixels)
        if self._offset_shft is not None:
            out_phi_shft -= cp.array(self._offset_shft)
        out_phi  = cp.fft.ifftshift(out_phi_shft)[
            (self._fft_width-self._pup_width)//2:(self._fft_width+self._pup_width)//2,
            (self._fft_width-self._pup_width)//2:(self._fft_width+self._pup_width)//2
                ]
        out_phi *= self._wavelength/(2*cp.pi)
        if hasattr(out_phi,"get"):
            return out_phi.get()
        else:
            return out_phi



try:
    import cupy as cp
except ImportError as ie:
    Warning("No cupy available, using basic numpy")
    import numpy as cp

import numpy as np

from scipy.special import factorial
import scipy.linalg as la

class TaylorHModel:
    def __init__(self,order,nmeas,nstate):
        self._order  = order
        self._nmeas  = nmeas
        self._nstate = nstate
        self._einbcd = "abcdefghijklmnopqrstuvwxyz"
        self.dny     = []
        self.g       = None
        assert order >= 0        
        if order > 5:
            Warning("High order taylor estimate, are you sure?")

        for ni in range(order+1):
            self.dny.append(cp.zeros([nmeas]+[nstate]*ni))
        self._einstring_eval = ["a"+self._einbcd[1:ni+1]+"".join([","+x for x in self._einbcd[1:ni+1]])+"->a" for ni in range(self._order+1)]
        self._einstring_jac  = ["ab"+self._einbcd[2:ni+2]+"".join([","+x for x in self._einbcd[2:ni+2]])+"->ab" for ni in range(self._order+1)]
        self._einstring_hess = ["abc"+self._einbcd[3:ni+3]+"".join([","+x for x in self._einbcd[3:ni+3]])+"->abc" for ni in range(self._order+1)]
        self.tot_flux = 1.0

    def set_dny(self,dny,n):
        assert dny.shape==self.dny[n].shape
        if type(dny) is cp.ndarray:
            self.dny[n] = dny.copy()
        else:
            self.dny[n][:] = cp.array(dny)
    
    def set_tot_flux(self,flux):
        self.tot_flux = flux

    def eval(self,x,order=None):
        if order is None:
            order = self._order
        if type(x) is np.ndarray:
            CPU = True
            x = cp.array(x)
        elif type(x) is cp.ndarray:
            CPU = False
        else:
            raise TypeError("input x must be either numpy or cupy array")
        out = self.dny[0].copy()
        for ni in range(1,order+1):
            out += (1/factorial(ni))*cp.einsum(self._einstring_eval[ni],self.dny[ni],*([x]*ni))
        if CPU:
            return out.get()
        else:
            return out
    
    def jacobian(self,x,order=None): 
        if order is None:
            order = self._order
        if type(x) is np.ndarray:
            CPU = True
            x = cp.array(x)
        elif type(x) is cp.ndarray:
            CPU = False
        else:
            raise TypeError("input x must be either numpy or cupy array")
        out = self.dny[1].copy()
        for ni in range(1,order+1-1):
            out += (1/factorial(ni))*cp.einsum(self._einstring_jac[ni],self.dny[ni+1],*([x]*ni))
        if CPU:
            return out.get()
        else:
            return out
        
    def hessian(self,x,order=None): 
        if order is None:
            order = self._order
        if type(x) is np.ndarray:
            CPU = True
            x = cp.array(x)
        elif type(x) is cp.ndarray:
            CPU = False
        else:
            raise TypeError("input x must be either numpy or cupy array")
        out = self.dny[2].copy()
        for ni in range(1,order+1-2):
            out += (1/factorial(ni))*cp.einsum(self._einstring_hess[ni],self.dny[ni+2],*([x]*ni))
        if CPU:
            return out.get()
        else:
            return out

    def ceo_build_dnys(self,gmt,src,imgr):
        n_px_fft = imgr.DFT_osf*imgr.N_PX_PUPIL
        pup = src.amplitude.host()
        xx,yy = np.meshgrid(np.arange(pup.shape[0]),np.arange(pup.shape[0]))

        pup_mask = pup==1

        phi = pup.copy()

        dft_matrix = la.dft(n_px_fft)
        dft_matrix = np.concatenate([dft_matrix[n_px_fft//2:,:],dft_matrix[:n_px_fft//2,]],axis=0)
        dft_matrix = dft_matrix[:,:src.phase.shape[0]]
        dft_matrix = dft_matrix[(n_px_fft-imgr.N_PX_IMAGE)//2+1:(n_px_fft+imgr.N_PX_IMAGE)//2+1,:]
        dft_matrix = np.kron(dft_matrix,dft_matrix)
        dft_matrix = dft_matrix[:,pup_mask.flatten()]

        ~gmt

        mode_to_phase = np.zeros([pup_mask.sum(),self._nstate])

        state = gmt.state
        x = state["M1"]["Txyz"][:,2]
        x *= 0.0
        gmt^=state
        ~imgr
        +src
        
        for i in range(self._nstate):
            x *= 0.0
            delta = 1e-7
            x[i] = delta
            gmt^=state
            +src
            phi = src.phase.host()
            mode_to_phase[:,i] = phi[pup_mask]/delta

        wavelength = src.wavelength
        self.wavelength = wavelength

        g0 = np.exp((1j*2*np.pi*(xx+yy)/pup.shape[0]/4))[pup_mask]
        self.g = lambda x: np.exp((1j*2*np.pi)*(mode_to_phase@x/wavelength+g0))
        self.h_true = lambda x: np.abs(dft_matrix @ self.g(x))**2
        self._mode_to_phase = mode_to_phase        
        self._dft_matrix    = dft_matrix

        d0h = np.abs(dft_matrix @ g0)**2
        self.set_dny(d0h,0)
        if self._order < 1:
            return
        
        d1h = np.zeros([imgr.N_PX_IMAGE**2,self._nstate])
        for ell1 in range(self._nstate):
            m_ell_ = mode_to_phase[:,ell1]
            d1h[:,ell1] = 2*(2*np.pi/wavelength)*((-1j)*((dft_matrix@(g0*m_ell_))).conj()*(dft_matrix@g0)).real
        self.set_dny(d1h,1)
        if self._order < 2:
            return

        d2h = np.zeros([imgr.N_PX_IMAGE**2,self._nstate,self._nstate])
        for ell1 in range(self._nstate):
            for ell2 in range(ell1+1):
                m_ell_1 = mode_to_phase[:,ell1]
                m_ell_2 = mode_to_phase[:,ell2]
                tmp = 2*((-1j*2*np.pi/wavelength)**2*(
                            (dft_matrix@(g0*m_ell_1*m_ell_2)).conj()*(dft_matrix@g0)
                            - (dft_matrix@(g0*m_ell_1)).conj()*(dft_matrix@(g0*m_ell_2))
                        )).real
                d2h[:,ell1,ell2] = tmp
                d2h[:,ell2,ell1] = tmp
        self.set_dny(d2h,2)
        if self._order < 3:
            return
            
        d3h = np.zeros([imgr.N_PX_IMAGE**2,self._nstate,self._nstate,self._nstate])
        for ell1 in range(self._nstate):
            for ell2 in range(ell1+1):
                for ell3 in range(ell2+1):
                    m_ell_1 = mode_to_phase[:,ell1]
                    m_ell_2 = mode_to_phase[:,ell2]
                    m_ell_3 = mode_to_phase[:,ell3]
                    tmp = 2*((-1j*2*np.pi/wavelength)**3*(
                            (dft_matrix@(g0*m_ell_1*m_ell_2*m_ell_3)).conj()*(dft_matrix@g0)
                            - (dft_matrix@(g0*m_ell_1*m_ell_2)).conj()*(dft_matrix@(g0*m_ell_3))
                            - (dft_matrix@(g0*m_ell_1*m_ell_3)).conj()*(dft_matrix@(g0*m_ell_2))
                            - (dft_matrix@(g0*m_ell_2*m_ell_3)).conj()*(dft_matrix@(g0*m_ell_1))
                        )).real
                    d3h[:,ell1,ell2,ell3] = tmp
                    d3h[:,ell1,ell3,ell2] = tmp
                    d3h[:,ell2,ell1,ell3] = tmp
                    d3h[:,ell2,ell3,ell1] = tmp
                    d3h[:,ell3,ell1,ell2] = tmp
                    d3h[:,ell3,ell2,ell1] = tmp
        self.set_dny(d3h,3)
        if self._order < 4:
            return
        else:
            raise ValueError("maximum taylor order implemented is 3")
    
    def exact_jacobian(self,x):
        if type(x) is np.ndarray:
            CPU = True
            x = cp.array(x)
        elif type(x) is cp.ndarray:
            CPU = False
        else:
            raise TypeError("input x must be either numpy or cupy array")
        b = self._dft_matrix_cp*self.g_cp(x)[None,:]
        out = ((b @ self._mode_to_phase_cp) * (b.conj().sum(axis=1))[:,None]).imag
        out *= -(2*cp.pi/self.wavelength)*2 # not sure why negative here, should be positive surely
        if CPU:
            return out.get()
        else:
            return out
    
    def _exact_jacobian(self,x):
        # deprecated, will be deleted soon
        b = self._dft_matrix*self.g(x)[None,:]
        d1h = ((b @ self._mode_to_phase) * (b.conj().sum(axis=1))[:,None]).imag
        d1h *= -(2*cp.pi/self.wavelength)*2 # not sure why negative here, should be positive surely
        return d1h

    def compass_build_dnys(self,sup,nwfs,get_phase):
        """build dnys for a single WFS using a COMPASS WFS.

        Currently only works for a single full-aperture WFS. Extension to higher
        order WFSs is relatively straightforward, but not implemented.

        Parameters
        ----------
        sup : CompassSupervisor object
            The supervisor object
        nwfs : int
            The index of the WFS to use
        get_phase : function
            A function which returns the phase of the WFS in the 
        """
        offset = np.array(sup.wfs._wfs.d_wfs[nwfs].d_offsets)
        offset -= offset.mean()
        self.general_build_dnys(pup=sup.get_s_pupil(),
            im_width=sup.wfs.get_wfs_image(nwfs).shape[0],
            fft_width=sup.wfs._wfs.d_wfs[nwfs].nfft, get_phase=get_phase,
            offset=-offset,
            wavelength=sup.config.p_wfss[0].get_Lambda())

    def general_build_dnys(self,pup,im_width,fft_width,get_phase,offset,wavelength,
                            delta=1.0):
        """Build the partial derivative arrays for a general simulation.
        
        Parameters
        ----------
        pup : ndarray (pup_width,pup_width)
            The pupil mask.
        im_width : int
            The width of the image.
        fft_width : int
            The width of the fft.
        get_phase : callable
            A function that takes a state vector and returns the phase in the
            pupil plane.
        delta : float
            The step size for the mode-to-phase matrix generation. I.e., if the
            function is a purely linear combination, then delta doesn't matter.
        """

        pup_width = pup.shape[0]
        nstate = self._nstate
        nmeas = self._nmeas

        dft_matrix = la.dft(fft_width,scale=None)
        dft_matrix = np.concatenate([dft_matrix[fft_width//2:,:],dft_matrix[:fft_width//2,]],axis=0)
        dft_matrix = dft_matrix[:,:pup_width]
        dft_matrix = dft_matrix[fft_width//2-im_width//2:fft_width//2+im_width//2,:]
        dft_matrix = np.kron(dft_matrix,dft_matrix)
        dft_matrix = dft_matrix[:,(pup==1).flatten()]

        mode_to_phase = np.zeros([(pup==1).sum(),nstate])

        x = np.zeros(nstate)        
        for i in range(nstate):
            x *= 0.0
            x[i] = delta
            phi = get_phase(x)[pup==1]
            mode_to_phase[:,i] = phi/delta

        self.wavelength = wavelength

        self._offset        = offset[pup==1]
        self._mode_to_phase = mode_to_phase
        self._dft_matrix    = dft_matrix
        self._offset_cp        = cp.array(offset[pup==1])
        self._mode_to_phase_cp = cp.array(mode_to_phase)
        self._dft_matrix_cp    = cp.array(dft_matrix)
        self.g = lambda x: np.exp(1j*(2*np.pi*(self._mode_to_phase@x/self.wavelength)+self._offset))
        self.g_cp = lambda x: cp.exp(1j*(2*cp.pi*(self._mode_to_phase_cp@x/self.wavelength)+self._offset_cp))
        self.h_true = lambda x: np.abs(self._dft_matrix @ self.g(x))**2
        self.h_true_cp = lambda x: cp.abs(self._dft_matrix_cp @ self.g_cp(x))**2
        g0 = self.g(x*0)

        d0h = np.abs(dft_matrix @ g0)**2
        self.set_dny(d0h,0)
        if self._order < 1:
            return
        
        d1h = np.zeros([nmeas,nstate])
        for ell1 in range(nstate):
            m_ell_ = mode_to_phase[:,ell1]
            d1h[:,ell1] = 2*(2*np.pi/self.wavelength)*((-1j)*((dft_matrix@(g0*m_ell_))).conj()*(dft_matrix@g0)).real
        self.set_dny(d1h,1)
        if self._order < 2:
            return

        d2h = np.zeros([nmeas,nstate,nstate])
        for ell1 in range(nstate):
            for ell2 in range(ell1+1):
                m_ell_1 = mode_to_phase[:,ell1]
                m_ell_2 = mode_to_phase[:,ell2]
                tmp = 2*((-1j*2*np.pi/self.wavelength)**2*(
                            (dft_matrix@(g0*m_ell_1*m_ell_2)).conj()*(dft_matrix@g0)
                            - (dft_matrix@(g0*m_ell_1)).conj()*(dft_matrix@(g0*m_ell_2))
                        )).real
                d2h[:,ell1,ell2] = tmp
                d2h[:,ell2,ell1] = tmp
        self.set_dny(d2h,2)
        if self._order < 3:
            return
            
        d3h = np.zeros([nmeas,nstate,nstate,nstate])
        for ell1 in range(nstate):
            for ell2 in range(ell1+1):
                for ell3 in range(ell2+1):
                    m_ell_1 = mode_to_phase[:,ell1]
                    m_ell_2 = mode_to_phase[:,ell2]
                    m_ell_3 = mode_to_phase[:,ell3]
                    tmp = 2*((-1j*2*np.pi/self.wavelength)**3*(
                            (dft_matrix@(g0*m_ell_1*m_ell_2*m_ell_3)).conj()*(dft_matrix@g0)
                            - (dft_matrix@(g0*m_ell_1*m_ell_2)).conj()*(dft_matrix@(g0*m_ell_3))
                            - (dft_matrix@(g0*m_ell_1*m_ell_3)).conj()*(dft_matrix@(g0*m_ell_2))
                            - (dft_matrix@(g0*m_ell_2*m_ell_3)).conj()*(dft_matrix@(g0*m_ell_1))
                        )).real
                    d3h[:,ell1,ell2,ell3] = tmp
                    d3h[:,ell1,ell3,ell2] = tmp
                    d3h[:,ell2,ell1,ell3] = tmp
                    d3h[:,ell2,ell3,ell1] = tmp
                    d3h[:,ell3,ell1,ell2] = tmp
                    d3h[:,ell3,ell2,ell1] = tmp
        self.set_dny(d3h,3)
        if self._order < 4:
            return
        else:
            raise ValueError("maximum taylor order implemented is 3")
    
    def _general_build_dnys(self,pup,im_width,fft_width,get_phase,offset,wavelength,
                            delta=1.0):
        # experimental, do not use. Goal is to build dnys for an arbitrary order,
        # but it's difficult to code up. We have the equations, but the implementation
        # is another thing.
        pup_width = pup.shape[0]
        nstate = self._nstate
        nmeas = self._nmeas

        dft_matrix = la.dft(fft_width,scale=None)
        dft_matrix = np.concatenate([dft_matrix[fft_width//2:,:],dft_matrix[:fft_width//2,]],axis=0)
        dft_matrix = dft_matrix[:,:pup_width]
        dft_matrix = dft_matrix[fft_width//2-im_width//2:fft_width//2+im_width//2,:]
        dft_matrix = np.kron(dft_matrix,dft_matrix)
        dft_matrix = dft_matrix[:,(pup==1).flatten()]

        mode_to_phase = np.zeros([(pup==1).sum(),nstate])

        x = np.zeros(nstate)        
        for i in range(nstate):
            x *= 0.0
            x[i] = delta
            phi = get_phase(x)[pup==1]
            mode_to_phase[:,i] = phi/delta

        self.wavelength = wavelength

        self.g = lambda x: np.exp(1j*(2*np.pi*(mode_to_phase@x/self.wavelength)+offset[pup==1]))
        self.h_true = lambda x: np.abs(self._dft_matrix @ self.g(x))**2
        g0 = self.g(x*0)
        self._mode_to_phase = mode_to_phase
        self._dft_matrix    = dft_matrix

        d0h = np.abs(self.dft_matrix @ g0)**2
        self.set_dny(d0h,0)
        if self._order < 1:
            return
        print("d0 done")
        import time
        t1 = time.time()

        b = self.dft_matrix*g0[None,:]
        d1h = ((b @ self.mode_to_phase) * (b.conj().sum(axis=1))[:,None]).imag
        d1h *= -(2*np.pi/self.wavelength)*2 # not sure why negative here, should be positive surely

        print("d1 done")
        print(f"Time for d1h: {time.time()-t1:0.2f} sec")

        self.set_dny(d1h,1)
        if self._order < 2:
            return
        
        d2h = np.zeros([self._nmeas,self._nstate,self._nstate])
        for ell1 in range(self._nstate):
            for ell2 in range(ell1+1):
                m_ell_1 = mode_to_phase[:,ell1]
                m_ell_2 = mode_to_phase[:,ell2]
                tmp = 2*((-1j*2*np.pi/self.wavelength)**2*(
                            (dft_matrix@(g0*m_ell_1*m_ell_2)).conj()*(dft_matrix@g0)
                            - (dft_matrix@(g0*m_ell_1)).conj()*(dft_matrix@(g0*m_ell_2))
                        )).real
                d2h[:,ell1,ell2] = tmp
                d2h[:,ell2,ell1] = tmp
        self.set_dny(d2h,2)
        if self._order < 3:
            return
        print("d2 done")
            
        d3h = np.zeros([self._nmeas,self._nstate,self._nstate,self._nstate])
        for ell1 in range(self._nstate):
            for ell2 in range(ell1+1):
                for ell3 in range(ell2+1):
                    m_ell_1 = mode_to_phase[:,ell1]
                    m_ell_2 = mode_to_phase[:,ell2]
                    m_ell_3 = mode_to_phase[:,ell3]
                    tmp = 2*((-1j*2*np.pi/self.wavelength)**3*(
                            (dft_matrix@(g0*m_ell_1*m_ell_2*m_ell_3)).conj()*(dft_matrix@g0)
                            - (dft_matrix@(g0*m_ell_1*m_ell_2)).conj()*(dft_matrix@(g0*m_ell_3))
                            - (dft_matrix@(g0*m_ell_1*m_ell_3)).conj()*(dft_matrix@(g0*m_ell_2))
                            - (dft_matrix@(g0*m_ell_2*m_ell_3)).conj()*(dft_matrix@(g0*m_ell_1))
                        )).real
                    d3h[:,ell1,ell2,ell3] = tmp
                    d3h[:,ell1,ell3,ell2] = tmp
                    d3h[:,ell2,ell1,ell3] = tmp
                    d3h[:,ell2,ell3,ell1] = tmp
                    d3h[:,ell3,ell1,ell2] = tmp
                    d3h[:,ell3,ell2,ell1] = tmp
        self.set_dny(d3h,3)
        print("d3 done")
        if self._order < 4:
            return
        else:
            raise ValueError("maximum taylor order implemented is 3")
    